/* HtmlFile.java 	- (c) Ilpo Jrvinen (ijjarvin@cs.helsinki.fi)
 *
 * Ohjelmoinnin harjoitusty kevt 2000
 *
 * -tabsize 4
 */
 
/**
 * HtmlFile
 * @author Ilpo Jrvinen <ijjarvin@cs.helsinki.fi>
 * @version 0.87 04.04.2000
 */
 
/*
 * This class uses HTML parsing technology from Arthur Do Consulting.
 */


import java.io.*;				// File & FileReader
import java.util.*;				// Vector & Enumeration
import adc.parser.*;			// the HTML parser

 
public class HtmlFile extends Timer {

	/* 
	 * all strings used in class
	 */
	public static final String	EMPTY 			= "";
	public static final String	FILE_INFO 		= ": ";
	public static final String	FILE_DONE 		= ": Done!";
//	public static final String	FILE_TIMEDOUT 	= ": Timedout!";	// debug  
	public static final String	FILE_MISSING 	= ": File not found!";
	public static final String	FILE_NOREAD 	= ": No read permission!";
	public static final String	FILE_NOWRITE 	= ": No write permission, reporting broken links!";
	public static final String	PARSER_ERROR 	= ": HTML parser failed!";
	public static final String	FILE_TEMPERROR	= ": Tempfile creation failed!";
	public static final String	FILE_WRITEERROR = ": Write failed!";
	
	public static final String	REPORT_DEPRECATED 	= " - Link Deprecated";
	public static final String	REPORT_WORKING 		= " - Working";
	
	public static final String	WRITE_DEPRECATED 	= "Link Deprecated - ";
	
	private static final int	INTERVAL		= 1;					// secs
	private static final int	MINTIMEOUT		= 50;					// secs


/* ********************************************************************** */
	private String				name;							// html filename
	private Vector				urls;							// list of urls inserted to pool
	private boolean				done;							// done?
	private File				file;							// Object holder
	private boolean 			canOnlyReport;					// write permission not granted?
	private boolean				timedOut;						// timeout occured during inspection
	
/* ********************************************************************** */
	private UserInterface		userInterface;					// ui provides output methods.
	private LinkPool			linkPool;						// the pool of links
	
	private File				tempFile;						// copy of the original htmlfile, used in writemode

/************************************************************************
 * This constructor...
 *
 * @param pool	Pool where links found are to be inserted.
 * @param ui	Userinterface where output will be sent.
 */
	public HtmlFile( String name, LinkPool pool, UserInterface ui ) 
	{
		super( MINTIMEOUT, ui.getTimeout(), INTERVAL );			// Timer, encounting...
		
		/*
		 * Setup local variables.
		 */
		done = timedOut = false;
		this.userInterface = ui;
		this.linkPool = pool;
		this.name = name;
		this.urls = new Vector();
		
		file = new File( name );					// nulls should never get this far!
		
		checkFileAccess();							// do all read/write/exist checks
		
		this.start();								// start this thread! then return to caller...
	}



/***************************************************************************
 * This method checks if all necessary permission are granted and that the
 * file actually is on the disk.
 */
	private void checkFileAccess() 
	{
		if (!file.exists()) {							// handle not found...
			userInterface.println( this.name + FILE_MISSING );
			done = true;
			return;
		}
		
		if (!file.canRead()) {						// handle no read permissions...
			userInterface.println( this.name + FILE_NOREAD );
			done = true;
			return;
		}
		
		if (!file.canWrite()) {						// handle no write permissions
			canOnlyReport = true;
			/*
			 * In the case that user requested write but it cannot be accomplished, report it.
			 * It's not clear whether this should be done here because it gets displayed twice!
			 * But this allows quick aborting if user notices a problem!
			 * Removed! It just looks stupid!
			 */
//			if (userInterface.writeSelected()) {
//				userInterface.println( this.name + FILE_NOWRITE );
//			}
		}
	}  
	
	

/*****************************************************************************
 * This method parses the html-file, waits until the timer runs out or all
 * links are done. Then calls doResults for resultmaking.
 */
	public void run() 
	{
		if (done)									// check if finished.
			return;

		if (parseHtml() != 0) {						// parse the hypertext.
			done = true;							// failed! :(
			return;
		}
		
		boolean linksDone = false;

		/*
		 * wait until links finish or timer expires!
		 */
		while (!super.isExpired() && !linksDone) {
			/*
			 * Go through the urls and see if every one of them has finished.
			 */
			linksDone = true;						// hope for the best
			for (Enumeration e = urls.elements(); e.hasMoreElements(); ) {
				String temp = (String)e.nextElement();
				if (!linkPool.isLinkFinished( temp )) {
					linksDone = false;				// the hope wasn't based on facts :(
					break;							// useless to continue...
				}
			}
			if (!linksDone)
				super.run();						// Let the Timer to wait for a while
//			userInterface.print( "." );				// DEBUG!
		}
		
		if (super.isExpired())
			timedOut = true;						// set flag for reporting when timer has expired

		/*
		 * Go to the resultmaker...
		 */
		doResults();
		done = true;								// we're done!
	}

	


/**************************************************************************
 * This method is used in report and reportall result modes. It will 
 * create a String which is later sent to output.
 *
 * @param 	result	The StringBuffer into which the output is built.
 */
	private void reportResults( StringBuffer result )
	{
		String LEAD = "  ";
		boolean didOutput = false;
		/*
		 * go through all urls under "surveillance"...
		 */
		for (Enumeration e = urls.elements(); e.hasMoreElements(); ) {
			String temp = (String)e.nextElement();
			/*
			 * see if it is broken, and report the state...
			 */
			if (linkPool.isLinkBroken( temp )) {
				result.append(LEAD).append(temp).append(REPORT_DEPRECATED).append('\n');
				didOutput = true;

			/*
			 * If all links must be reported, print the OKs too... 
			 * Never prints any relative links! :( 
			 * This is OK because those are not verified either.
			 */
			} else if (userInterface.reportAllSelected()) {
				result.append(LEAD).append(temp).append(REPORT_WORKING).append('\n');
				didOutput = true;
			}
		}
		if (!didOutput)
			result.setLength(0);			// clear if no output was done!
		result.append(name).append(FILE_DONE);
	}
	

/*****************************************************************************************
 * Method copies inFile into other file, file extension will be ".lvtmp". If temp file 
 * already exists it will not be overwritten but the method returns null instead. JDK 1.2+
 * would have method which would find unique filename. No reason to fix JAVA idiotism, 
 * just get the later version and recompile.
 *
 * @return <b>File</b> if one was copied successfully, null otherwise.
 */	
	private File copyFile( File inFile )
	{
		BufferedReader	in;
		PrintWriter		out;
		String			line;
		File			tempFile;				// Warning! shadows an instance variable!!!

		try {
//			tempFile = new File.createTempFile( "html", ".lvtmp" );		// 1.2+ :(

			tempFile = new File( inFile.getName() + ".lvtmp" );
			if (tempFile.exists()) {
				return null;					// file creation failed... :(
			}
			
			/*
			 * Open the streams and do the duplicate.
			 */
			in = new BufferedReader( new FileReader( inFile ) );
			out = new PrintWriter( new FileWriter( tempFile ), true );
			while ((line = in.readLine()) != null) {
				out.println( line );
			}
			out.flush();						// write do disk
			out.close();
		} catch (Exception e) {
			return null;
		}
		return tempFile;
	}


/**************************************************************************
 * Writes the results to html-file, caller should check permissions.
 * This method is sligthly "broken", the output of the parser could be 
 * used but it would "break" the original structure of the html-file and
 * therefore is not desirable. Instead of this the url "keys" are searched
 * inside of every "A"-tag, if such text is some point before the actual
 * href a false match is can found. The problem is very unlikely to occur
 * in "good" html source but it is very easy to construct such situation!
 * because the false match _requires_ the "-marks around the address...
 * I don't know HTML well enough to say if this ever occurs?
 *
 * @param	result	The StringBuffer into which the screen output is built.
 */
	private void writeResults( StringBuffer result )
	{
		/*
		 * Lock the Runtime! Prevents anything from exiting while disk work
		 * is done here... Doesn't guarantee a trap for SIGNAL (SIGTERM and 
		 * SIGQUIT) based terminations too. :( The main however will have 
		 * to wait! This should never happend, but it's still better to hold
		 * the lock anyway...
		 */
		synchronized (Runtime.getRuntime()) {

			Enumeration		url = urls.elements();
			
			boolean			inTag = false;				// tag is open
			boolean			tagFirstLetter = false;		// tag's first letter has been examined
			boolean			aTag = false;				// it was <a ...>-tag
			
			String			address = null;
			boolean			foundAddress = false;		// address was found inside the tag?
			boolean			deprecated = false;			// address is deprecated?
			
			BufferedReader	in;
			String			line;
			PrintWriter		out;
			
			/*
			 * Copy to a temp file, if this fails, forget the whole op!
			 * Report an error instead.
			 */
			tempFile = copyFile( file );
			if (tempFile == null) {
				result.append( FILE_TEMPERROR );
				return;
			}

			try {
				/*
				 * Open the streams, if this fails Exception will occur...
				 */
				in = new BufferedReader( new FileReader( tempFile ) );
				out = new PrintWriter( new FileWriter( file ), true );

				/*
				 * read the whole input, one line at a time
				 */
				while ((line = in.readLine()) != null) {
					int startIndex = 0, endIndex;
					/*
					 * Do the loop if there is a tag "starter" on the line or tag is "open" from previous line...
					 */
					while ((!inTag && line.indexOf('<', startIndex) >= 0) || inTag) {
						/*
						 * Calculate tag begin and end
						 */
						if (!inTag) {
							startIndex = line.indexOf('<', startIndex ) + 1;
							inTag = true;
							tagFirstLetter = false;
						}
						
						/*
						 * Search the type of tag.
						 */
						if (!tagFirstLetter) {
							for (int i = startIndex; i < line.length(); i++) {
								if (Character.isWhitespace(line.charAt(i)))
									continue;
								tagFirstLetter = true;
								aTag = false;
								/*
								 * check if it's "a"-tag, not "aX..."-tag where X is not a whitespace.
								 */
								if (Character.toUpperCase(line.charAt(i)) == 'A' && 
									(line.length() == i + 1  || Character.isWhitespace(line.charAt(i+1)) )) {
									aTag = true;
								}
								break;
							}
						}
						endIndex = line.indexOf('>', startIndex);
						if (endIndex < 0)
							endIndex = line.length();
						/*
						 * If there is more address to check and we're not in the middle of 
						 * checking one, pick the next one.
						 */
						if (!foundAddress && address == null && url.hasMoreElements())
							address = ((String)url.nextElement());
						
						/*
						 * Look for the address if it's valid.
						 */
						if (address != null && aTag) {
							int addrIndex = line.indexOf( "\""+address+"\"", startIndex );
							if (addrIndex < endIndex && addrIndex >= 0) {		// must be inside the tag.
								if (linkPool.isLinkBroken(address))				// was this address broken?
									deprecated = true;							// mark "it was!"
								address = null;
								foundAddress = true;
							}
						}
						/*
						 * the tag doesn't end on current line!
						 */
						if (endIndex == line.length())
							break;
						/*
						 * There was the address inside the tag, and tag terminator is on this line!
						 * Calculate the position of "Link Deprecated message" and either add or
						 * remove it.
						 */
						if (foundAddress) {
							boolean hasDeprecated = line.regionMatches( endIndex+1, WRITE_DEPRECATED, 0, WRITE_DEPRECATED.length() );
							if (deprecated && !hasDeprecated) {
								line = line.substring( 0, endIndex+1 ) + WRITE_DEPRECATED + line.substring( endIndex+1 );
							} else if (!deprecated && hasDeprecated) {
								line = line.substring( 0, endIndex+1 ) + line.substring( endIndex+1 + WRITE_DEPRECATED.length() );
							}
							deprecated = false;
							foundAddress = false;							// this address has been solved.
						}
						
						/*
						 * Look for the next tag, starting end of the last one...
						 */
						startIndex = endIndex;
						inTag = false;
					}
					out.println( line );									// write line to file.
				}
				
				out.flush();						// write to file
				out.close();						// close the stream
		
				tempFile.delete();
				tempFile = null;
				
			} catch (Exception e) {
				/*
				 * Lock the object: file renaming is being done and we DON'T want any finalizers during this period.
				 * Sadly enough ThreadDeath can release the lock but normal user cannot make it to happend.
				 * Obviously, the ThreadDeath could be caught... :)
				 */
				synchronized (this) {
					if (tempFile != null) {
						File temp = tempFile;
						tempFile = null;
						file.delete();
						temp.renameTo(file);				// may fail, but no good can be done if it does!
					}
				}
				result.append( FILE_WRITEERROR );
//				e.printStackTrace(System.err);		// DEBUG!!!!
				return;
			} // eot
		} // eos

		/*
		 * Everything went just fine. :)
		 */
		result.append( FILE_DONE );
		return;
	}


/*************************************************************************
 * Result multiplexer... "Go wherever the booleans takes you"... ;)
 */
	private void doResults() 
	{
		StringBuffer out = new StringBuffer( name );
		/*
		 * Write if requested and permissions are ok -> Writemode
		 */
		if (userInterface.writeSelected() && !canOnlyReport) {
			writeResults( out );
		} else {
			if (canOnlyReport && userInterface.writeSelected()) {
				out.append(FILE_NOWRITE).append('\n');					// write wasn't possible, tell to the user
			} else {
				out.append(FILE_INFO).append('\n');						// otherwise do "normal" message
			} 
			/*
			 * Reportmode
			 */
			reportResults( out );
		}
		userInterface.println( out.toString() );
	}


/***********************************************************************
 * Returns the status of this file, true if file is complete or error
 * which prevents additional processing has occured.
 */
	public boolean isFinished()				{ return done; }


/***********************************************************************
 * Searches all <a href...> tags from html file and insert them to pool.
 *
 * @return Negative error code if fails. 
 */
	private int parseHtml()
	{
		HtmlStreamTokenizer 	token;
		HtmlTag					tag = new HtmlTag();

		/*
		 * open the stream, if this fails we're done!
		 */
		try {
			token = new HtmlStreamTokenizer( new FileReader( file ) );
		} catch (FileNotFoundException e) {
			return -1;
		}

		/*
		 * try to search all links from stream
		 */
		try {
			while (token.nextToken() != HtmlStreamTokenizer.TT_EOF) {
				
				int ttype = token.getTokenType();
				
				/*
				 * only tags may have hrefs
				 */
				if (ttype == HtmlStreamTokenizer.TT_TAG) {
				
					try {
						/*
						 * verify tag type...
						 */
						token.parseTag(token.getStringValue(), tag );
						if (tag.getTagType() == HtmlTag.T_A && !tag.isEndTag() && tag.hasParam("href") ) {
							/*
							 * href was found, add to pool (and urls, if was added to pool)...
							 */
							if (linkPool.addLink( tag.getParam("href") )) {
								urls.addElement( tag.getParam("href") );
								super.add();							// add more time!
							}
						}
					/*
					 * Exception means html is broken or that input is not html at all -> should probably
					 * warn user if such condition is detected! can the result of ignorance be an endless loop?
					 * Only empty tag Exceptions are currently ignored. It would be nice to try to continue even
					 * when others occurs. But the parser probably can't handle that... :( And it's must be broken
					 * HTML anyway...
					 */
					} catch (HtmlException e) {
						if (e.getMessage().equals("parse empty tag"))
							continue;
						userInterface.println( name + PARSER_ERROR );
						return -4;
					}
				}
			}
			
		} catch (IOException e) {
			/*
			 * Mysterious error from HTML-parser, is _not_ our prob! "Rename" it...
			 */
			userInterface.println( name + PARSER_ERROR );
			return -2;
		}

		return 0;
	}


/***********************************************************************************
 * Compares this <b>HtmlFile</b> to another <b>Object</b>. The result is true if and 
 * only if the name of the html-filename matches with the given <b>String</b>.
 */
	public boolean equals( Object o )
	{
		if (o instanceof String) {
			if (o == null)				// can null get this far, is it instanceof anything?
				return false;			// NullPointerExceptions are NO, NO!!!
			return file.equals( new File((String)o) );		// File.equals handles Upper/Lowercase traps (OS-dependant)
		}
		return false;
	}


/***********************************************************************************
 * This method is run when instance dies, this ensures the tempfile deletion, which
 * is a "special-case" that should never happend unless program dies unexpectedly
 * (=Java virtual machine terminates. And the source of the termination wasn't the
 * program)
 */	
	public void finalize()
	{
		if (tempFile != null)
			tempFile.delete();
		tempFile = null;
	}
	

}	// eoc
