/* Link.java	- (c) Ilpo Jrvinen (ijjarvin@cs.helsinki.fi)
 *
 * Ohjelmoinnin harjoitusty kevt 2000
 *
 * All commercial use is prohibited!
 *
 * -tabsize 4
 */

/** 
 * Link
 * @author Ilpo Jrvinen <ijjarvin@cs.helsinki.fi>
 * @version 0.6a 04.04.2000
 */


import java.net.*;
import java.io.*;

/** 
 * Class Link tries to create a stream to specified destination. Many instances
 * can be run at parallar because all high latency operations are done in the
 * <b>run()</b> method.
 *
 * @see java.net
 * @see java.io
 */


public class Link extends Thread {

	/**
	 * Destination was found, or unknown protocol.
	 */
	public final int		LINKOK = 0;
	/**
	 * Not yet connected, exact state unknown.
	 */
	public final int		CONNECTING = -1;

	/**
	 * Specified hostname does not exist.
	 */
	public final int		HOSTERROR = -100;

	/**
	 * The specified host does not respond.
	 */
	public final int		PORTERROR = -200;

	/**
	 * Hostname was correct but specified file was not found.
	 */
	public final int		FILEERROR = -300;

	/**
	 * Hardware or operating system could not connect to specified destination.
	 */
	public final int		HWOSERROR = -400;

	/**
	 * The specified url was not qualified.
	 */
	public final int		MALFORMED = -500;

	/**
	 * There was no route to specified host, network could be temporarily (?) down.
	 */
	public final int		NOROUTE	  = -1000;


/* ********************************************************************************* */
	private int 					error;			// < 0 when non http error
	private URL						url;			// network object...
	private URLConnection			connection;		// ...holders


/**************************************************************************************
 * Creates an <b>URL</b> object from <b>String</b> destination. 
 *
 * @param	url		destination address
 * @see 			java.net.URL
 */
	public Link( String url ) 
	{
		error = CONNECTING;
		
		try {
			this.url = new URL( url );					// latency is acceptable, NO DNS query will be done here (checked)!
		} catch( MalformedURLException e ) {
			error = MALFORMED;							// catch faulty addresses
		}
		
		/*
		 * lower the priority, don't set to MIN_PRIO though... (=on some OSes that could
		 * mean 'run only when CPU is idle' which is not what we're looking for....
		 * This wasn't so good idea after all, linux Java Virtual Machine Vx.yy.zz failed
		 * to run program correctly when this was in use, 99%+ of cpu resources used, no
		 * work got done! Is the interface between jvm and kernel broken?
		 */
//		this.setPriority( (Thread.NORM_PRIORITY + Thread.MIN_PRIORITY) / 2 );

		this.start();									// Start the thread (this)
	}
	

/**************************************************************************************
 * Attempts to connect to specified destination, sets an error which can
 * be queryed by using method <b>getError()</b>, the user must use <b>isFinished()</b>
 * method to determinate whether this method has finished or not.
 *
 * @see				java.net.URLConnection
 * @see				java.net.HttpURLConnection
 */
	public void run () 
	{
		if (error != CONNECTING)									// check if already in known state?
			return;
			
		try {
			connection = this.url.openConnection();					// create the connection to target
			
			if (connection instanceof HttpURLConnection) {			// see if we estabilished an http connection?			
				HttpURLConnection http = (HttpURLConnection) connection;			// It was a http ->
				int code = http.getResponseCode();					// now find out if the targeted page is really there...
				switch (code) {
					
					case	HttpURLConnection.HTTP_OK:				// http-server told "everything ok!"
						error = LINKOK;
						break;
					
					default:
						/* 
						 * something is not correct with the speficified target...
						 * CONSIDERME?!? should these codes be mapped to < 0 values, maybe do it in getError()
						 * because otherwise the "real" cause of troubles will remain uncertain...
						 * No error code will ever be used in current version of the software ->
						 * no mapping is done!
						 */
						error = code;
				}
			} else {
				/*
				 * It wasn't a http, just check if stream can be established. Exception will be thrown if this fails!
				 * If stream can be established, assume that the target is OK!
				 */
				InputStream holder = connection.getInputStream();
				error = LINKOK;
			}
			
		/*
		 * all Exceptions will just set error to correct value, 
		 * every Exception caused by operations above will be handled down here.
		 */
		} catch (NoRouteToHostException e) {
			error = NOROUTE;
		} catch (UnknownHostException e) {
			error = HOSTERROR;
		} catch (ConnectException e) {
			error = PORTERROR;
		} catch (FileNotFoundException e) {
			error = FILEERROR;
		} catch (UnknownServiceException e) {
			error = LINKOK;								// it's not clear how this should be handled, so assume it's ok!
		} catch (IOException e) {
			/*
			 * I got FtpProtocolException: LIST
			 * from ftp.funet.fi?
			 * should it be assumed to be an error? This "feature" or "bug" is unreproduceable due nature of internet.
			 */
			error = HWOSERROR;
		}
		/*
		 * Try to free some resources! We don't need these objects!
		 * Hopefully Java is wise enough and releases the ports too! (=not! :( )
		 * Otherwise a bottlenecks will occur in OS-level code...
		 */
		connection = null;
		url = null;

		/*
		 * Java hack...
		 */
		System.gc();									// run garbage collector for released objects!
		this.yield();									// let finalizers run (=this thread sleeps a while)
	}

/****************************************************************************
 * Returns the status of the connection.
 *
 * @return		true if connection has not yet been established.
 */
	public boolean isFinished()		{ return error != CONNECTING; }

	
/****************************************************************************
 * Returns the status of the destination. This information will not be
 * reliable as long as <b>isFinished()</b> returns false.
 *
 * @see			#isFinished()
 * @return		false if link is working.
 */
	public boolean isBroken() 		{ return error != LINKOK; }


/****************************************************************************
 * Returns the status of the destination. This information will not be
 * reliable as long as <b>isFinished()</b> returns false.
 *
 * @see			#isFinished()
 * @return		true if the destinated page has moved to new address.
 */
	public boolean isMoved() 		{ return error == HttpURLConnection.HTTP_MOVED_PERM; }


/****************************************************************************
 * Returns numerical representation of error code. 
 * Positive value means HTTP-specific error, negative for any other protocol.
 *
 * @return		the errorcode, positive for http-specific errors.
 */
	public int getError()			{ return error; }


/****************************************************************************
 * Returns <b>String</b> representation of specified Link and its state.
 * USED ONLY FOR DEBUGGING! Advanced version might take advantage of this
 * and report more specific information to 'guru'... ;)
 */
	public String toString()
	{ 
		if (url == null)
			return "";
		StringBuffer s = new StringBuffer(url.toString()).append(" - ");
		switch (error) {
			case NOROUTE: s.append( "host not reachable. (temporarily?)" ); break;
			case HOSTERROR: s.append( "host not found" ); break;
			case PORTERROR: s.append( "destination port did not respond." ); break;
			case FILEERROR: s.append( "file not found" ); break;
			case HWOSERROR: s.append( "hardware or OS did something bad." ); break;
			case LINKOK: s.append( "up and running." ); break;
			case MALFORMED: s.append( "address given is invalid." ); break;
			case CONNECTING: s.append( "still trying..." ); break;
			default: 
				if (error > 0) {
					s.append( "protocol code " + error + "." );
				} else {
					s.append( "unknown error." ); 
				}
				break;
		}
		return s.toString(); 
	}

}


